/**
 * This file is part of the "Deniable Encryption" release in support of the
 * Blackhat Asia 2014 talk:
 *    A. Trachtenberg, "Say It Ain't So - An Implementation of Deniable Encryption".
 * 
 * The code in this file may code under patent and is released under the
 * following agreement:
 * http://nislab.bu.edu/wp-content/uploads/2014/03/Patented-Software-Release-Agreement.pdf
 */

/* 
 * File:   DeBruijnGraph.h
 * Author: Ari Trachtenberg
 * Implements some functions on DeBruijn graphs with much less than MAX_LONG edges.
 *   (more precisely, the graph must have 2*edges^2 < MAX_LONG)
 *
 * Created on November 11, 2013, 2:05 PM
 */

#ifndef DEBRUIJNGRAPH_H
#define	DEBRUIJNGRAPH_H

// includes
#include <string>
#include <set>
#include <map>

using namespace std;

class DeBruijnGraph {
    typedef std::set<string> vtype; // a vertex
    typedef string etype_v1;   // the first first in an edge
    typedef string etype_v2;   // the second vertex in an edge
    typedef long etype_weight;  // the weight of an edge
    typedef etype_v1 etype_first; // first component in an edge = first vertex
    typedef multimap<etype_v2, etype_weight> etype_second; // second component in an edge = second vertex -> weight
    typedef map<etype_first, etype_second> etype; // edge type

public:
    /**
     * Creates a DeBruijn multi-graph representation of a shingle set.
     * Each edge in the graph is associated with an integer weight.
     * @param shingles The shingles with which to create the graph.
     */
    DeBruijnGraph(multiset<string> shingles);

    DeBruijnGraph(const DeBruijnGraph& orig);

    // Graph query

    /**
     * Produce a human-readable version of the graph, expressed as a string.
     * @return A human-readable version of the graph.
     */
    string print();

    /**
     * Performs a Depth-First-Search of the graph, starting at node <start>
     *    Where there are several options for edges to traverse, chooses the edge with highest weight.
     * @modifies edge weights are clobbered during this algorithm to avoid additional bookkeeping
     * @require The graph must have an Eulerian cycle starting with <start>
     * @return A string representing the characters visited during the traversal.
     */
    string DFS(etype_v1 start);
    
    // Graph manipulation
    /**
     * Adds a vertex to the graph.
     * @param vertex The label of the vertex.  All graph labels must be unique.
     * @return false If vertex could not be inserted into the graph (e.g. if a vertex with the same label already exists).
     */
    bool addVert(etype_v1 vertex);

    /**
     * Adds an edge in the graph from vertex <v1> to vertex <v2> with weight <weight>
     * @param v1 The label of the first vertex.
     * @param v2 The label of the second vertex.
     * @param weight The weight of the edge.
     * @requires Cannot have more than MAX_LONG edges in the graph
     */
    void addEdge(etype_v1 v1, etype_v2 v2, etype_weight weight);
    
    // Static methods
    /**
     * Produce a shingling of a <text> with shingles of length <shingleLen>
     * @param text The text to shingle.
     * @param shingleLen The length of each shingle.
     * @return A multiset of the shingles corresponding to the text.
     */
    static multiset<string> shingle(string text, int shingleLen);
    
    /*
     * Checks whether the length <shingleLen> shingling of the given text <text>
     * can be uniquely decodable.  Uses the Kontorovich-Trachtenberg algorithm from ITA 2013.
     * 
     * @param shingleLen The shingle length of the text
     * @return true iff the length <shingleLen> shingling of text is uniquely decodable
     */
    static bool CheckUniqDecod(string text, int shingleLen);

    /**
     * Decodes a multiset of shingles into a string (i.e. finds an Eulerian cycle in the DeBruijn Graph).
     * @require The multiset <shingles> must be uniquely decodable.
     * @require shingles much each have length >0.
     * @param   shingles A multiset of shingles to decode.
     * @param   start The shingle with which to start the decoding
     * @return  The string whose shingles are represented by <shingles>, starting at <start>
     */
    static string Decode(multiset<string> shingles, string start);

private:
    // constants
    const static etype_weight INIT_WEIGHT = 1; /** the initial weight of an edge */
    
    // fields
    string startVer; // The start vertex of the graph
    string currVer; // The current vertex in a traversal of the graph
    vtype Vert; // the Vertices
    etype Edge; // an edge maps on vertex to another vertex with an integer weight
    long edgeID; // an ID of an edge as it is traversed
    long totalEdges;// the total number of edges in the graph
    
    // methods
     /**
     * Performs a Depth-First-Search of the graph, updating each edge weight to a number corresponding to the visiting order
     * @modifies edge weights
     * @param start The start node for the search.
     */
    void DFSupdate(etype_v1 start);
    void DFSupdateVisit(etype_v1 start); /** helper procedure for DFSupdate - searches subtree rooted at <start> */
    string DFSvisit(etype_v1 start);       /** helper procedure for DFS - searches subtree rooted at <start> */
};

#endif	/* DEBRIUJNGRAPH_H */


