/**
 * This file is part of the "Deniable Encryption" release in support of the
 * Blackhat Asia 2014 talk:
 *    A. Trachtenberg, "Say It Ain't So - An Implementation of Deniable Encryption".
 * 
 * The code in this file may code under patent and is released under the
 * following agreement:
 * http://nislab.bu.edu/wp-content/uploads/2014/03/Patented-Software-Release-Agreement.pdf
 */

/* 
 * File:   Deniable.h
 * Author: Ari Trachtenberg
 *
 * Created on February 7, 2014, 12:00 PM
 */

#ifndef DENIABLE_H
#define	DENIABLE_H

#include <vector>
#include <string>
#include <set>
#include <iostream>

using namespace std;

class Deniable {
public:

    // CONSTANTS
    static const int CHAR_THRESHOLD = 5; /** The minimum ASCII value in a plaintext. */
    // ... These constants should be values < CHAR_THRESHOLD
    static const char DELIMITER_NULL = '\0'; // NULL character - a delimiter assumed not to exist in the input texts
    static const char DELIMITER = '\1'; // a delimiter assumed not to exist in the input texts
    static const int MIN_SEC = 50; // minimum security bits; brute-force attack needs to sort through 2^MIN_SEC possibilities to break encryption

    // MEMBER FUNCTIONS
    /**
     * Encrypts a number of plaintexts into one ciphertext and a number of keys.
     * The k-th key, when applied to the ciphertext, produces the n-th plaintex.
     * 
     * @param plaintexts A (non-empty) vector of plaintexts to be encrypted.  Texts should
     * not have any characters with ASCII value < CHAR_THRESHOLD.
     * 
     * @param perr The maximum probability that an incorrect key will not produce a decoding failure
     * (this is the probability of error of the underlying sync)
     * 
     * @param sec A security paramenter:
     * When set to -1, the security parameter is automagically determined by the program
     * When set to 0, this scheme degenerates to a one-time pad - full security, long key
     * When set high, this scheme produces weaker security, but a shorter key
     * 
     * @return A pair.  The first item of the pair is the ciphertext, the second
     * item is a vector of keys, the i-th of which may be used to decode the i-th
     * plaintext.
     */
    static pair< string, vector<string> > encrypt(const vector<string> &plaintexts, double perr = 1E-10, int sec = -1);

    /**
     * Decrypts a given ciphertext with a given key.
     * @param key The key with which to decode.
     * @param ciphertext The name of the ciphertext file to decode.
     * @return A plaintext corresponding to the given ciphertext and key, if decryption succeeds.
     * @throws a DecryptionFailure exception if decoding was unsuccessful.
     */
    static string decrypt(const string key, const string ciphertext);

    // CLASSES

    /**
     * Thrown upon a decryption failure.
     */
    class DecryptionFailure : public exception {
    public:

        DecryptionFailure(string theKey, string theCipher) : key(theKey), ciphertext(theCipher) {
        }

        virtual ~DecryptionFailure() throw () { } // default destructor
        string key; // the key that caused the exception
        string ciphertext; // the name of the ciphertext file that caused the exception
    };

    /**
     * Thrown upon an encryption failure
     */
    class EncryptionFailure : public exception {
    public:
        EncryptionFailure(string theReason) : reason(theReason) {}
        virtual ~EncryptionFailure() throw () { } // default destructor
        string reason; /** The reason for the encryption failure. */
    };
    
    // HELPER FUNCTIONS
    /**
     * Reads a file into a string and returns the string.
     * @param fileName The file name to read.
     * @return The contents of the file, as a string.
     */
    static string readFile(string fileName);

    /**
     * Writes <contents> to a file with name <fileName>
     * @param fileName The name of the file to which tow rite.
     * @param contents What to write to the file.
     */
    static void writeFile(string fileName, string contents);

private:

    // HELPER FUNCTIONS
    /**
     * Helper procedure for shortestShingle()
     * @param plaintext The text to search
     * @param lower A lower bound on the shingle length needed to uniquely shingle <plaintext>
     * @param upper An upper bound on the shingle length needed to uniquely shingle <plaintext>
     * @return 
     */
    static int shortestShingleHelp(const string plaintext, int lower, int upper);

    /**
     * Finds the shortest shingle length under which the given plaintext is uniquely decodable.
     * Uses a binary search to find the right shingle length <shingleLen>.
     * @param plaintext The plaintext to test
     * @return The shingle length.
     */
    static int shortestShingle(const string plaintext);

    /**
     * Runs one-way CPISync to the multiset <shingles> to enable determination of up to <maxDiff>
     * differences with a probability of error <perr>.
     * @param shingles The multiset of shingles whose characteristic polynomial needs to be evaluated.
     * @param maxDiff The maximum number of differences that can be determined with the output provided.
     * @param bb The number of bits used to represent a set element
     * @param perr The maximum probability of error of reconciling up to maxDiff differences against
     *      the output provided.
     * @require <shingles> contains at least one shingle.
     * @require All shingles are of the same length, and consist of char's.
     * @return <maxDiff>\n<perr>\n<The base64 encoded output of CPISync>
     */
    static string evalCPI(multiset<string> shingles, int maxDiff, int bb, double perr);

    /**
     * Synchronize given multiset of data A with the CPI hash of a different multiset B, returning B.
     * @param hashContents The output of a oneway CPIsync of multiset B of strings.
     *        The CPIsync must have been done with <maxDiff> differences and probability of error <perr>.
     * @param myData The contents of multiset A of strings.
     * @param maxDiff The maximum number of differences between A and B, as included in the construction of <hashContents>.
     * @param bb The number of bits per set element.
     * @param perr The maximum probability of error included in the construction of <hashContents>.
     * @require All strings in the multisets must be the same length.
     * @require There must be at least one string in myData.
     * @return The contents of multiset B.
     */
    static multiset<string> syncCPI(string hashContents, const multiset<string> myData, int maxDiff, int bb, double perr);

    /**
     * Attempts to decode string <encoded> with key <key>.
     * @param encoded The encoded text.
     * @param key The key being used for decoding.
     * @require key length must be the same as the shingle length used to produce <encoded>
     * @return The decoded output.
     */
    static string decodeWithKey(string key, string encoded);

    /**
     * (Greedily) combine strings in a given multiset of shingles where the end of the first
     * string matches the beginning of another string.  The current implementation is worst-case
     * quadratic time in the number of shingles.
     * @param shingles The shingles to attempt to combine
     * @return A collection of shingles, some of which may have been combined
     */
    static multiset<string> compress(const multiset<string> shingles);

    /**
     * Computes a key for a given collection of shingles (the ciphertext is computed elsewhere).
     * @param shingleLen The common length of each shingle.
     * @param shingles The shingles to include in the key.
     * @requires Maintains the assumption that \0 does not appear within any shingle
     * @return A string representing a valid key for the given parameters.
     */
    static string computeKey(int shingleLen, const multiset<string> &shingles);
};

#endif	/* DENIABLE_H */