/*
 * This code is part of the CPISync project developed at Boston University:
 *      {@link http://nislabcpi.bu.edu/cpisync}
 * 
 * The code, explanation, references, API, and a demo can be found on this
 * web page.  The main theoretical basis for the approach can be found at
 * {@link http://ipsit.bu.edu/documents/ieee-it3-web.pdf}
 *   Y. Minsky, A. Trachtenberg, and R. Zippel,
 *   "Set Reconciliation with Nearly Optimal Communication Complexity",
 *   IEEE Transactions on Information Theory, 49:9.
 * 
 * Elements of the CPISync project code have been worked on, at various points, by:
 * @author Ari Trachtenberg
 * @author Sachin Agarwal 
 * @author Paul Varghese
 * @author Jiaxi Jin
 * @author Jie Meng
 * @author Alexander Smirnov
 * @version 2.01, 11/20/2013
 */

// GenSync.h -- The general synchronization framework
#ifndef GEN_SYNC_H
#define GEN_SYNC_H

// standard libraries
#include <list>
#include <string>

// namespaces
using namespace std;

// forward declarations
class DataObject;
class Communicant;
class SyncMethod;

/**
 * Implements a data structure for storing sets of data
 * in a manner that is designed for efficient synchronization.
 * This is a generic container that is subclassed by different types of
 * synchronization options.
 * 
 * Elements of this entire project have been worked on, at various points, by:
 * @author Ari Trachtenberg
 * @author Jie Meng
 * @author Jiaxi Jin
 * @author Alexander Smirnov
 * @author Sachin Agarwal
 * @version 2.00, 7/19/2011

 */
class GenSync {
public:
    /**
     * Specific GenSync constructor
     * @param clist      The list of other GenSync's with whom this data structure might
     *                      be synchronized.
     * @param mlist      The list of synchronization methods that this GenSync
     *                      should be prepared to use.  The order of these methods
     *                      is significant.
     * @param data       The initial data with which to populate the data structure.  The data is added element by element
     *                      so that synchronization method metadata can be properly maintained.
     * 
     */
    GenSync(const list<Communicant*> clist, const list<SyncMethod*> mlist, const list<DataObject*> data);

    /**
     * @param clist      The list of other GenSync's with whom this data structure might
     *                      be synchronized.
     * @param mlist      The list of synchronization methods that this GenSync
     *                      should be prepared to use.  The order of these methods
     *                      is significant.
     * @param fileName   The name of a file from which to read (line by line) initial elements of
     *                   this data structure.  As elements are added to this data structure, they
     *                   are also stored in the file.
     */
    GenSync(const list<Communicant*> clist, const list<SyncMethod*> mlist, string fileName = "");


    // DATA MANIPULATION
    /**
     * Adds a new datum into the existing GenSync data structure
     * @param newDatum   The datum to be added
     * %R:  newDatum cannot have size larger than a long
     * %M:  If a file is associated with this object, then updates are stored in that file.
     */
    void addElem(DataObject* newDatum);
    
    /**
     * Adds a new datum into the existing GenSync data structure
     * @param newDatum The datum to be added ... must be of a type compatible with
     * the global toStr templated function.
     * %R:  newDatum cannot have size larger than a long
     * %M:  If a file is associated with this object, then updates are stored in that file.
     *      *    */
    template <typename T>
    void addElem(T* newDatum);

    /**
     * Deletes a given element from the GenSync data structure
     * Not currently implemented.
     * @unimplemented
     */
    void delElem(DataObject* newDatum);

    /**
     * @return a list of pointers to the elements stored in the data structure
     */// get a data object element from data list
    const list<DataObject *> dumpElements();


    // COMMUNICANT MANIPULATION
    /* Communicants are entities that can communicate [and thus sync] with this GenSync object.
     *  They must each be registered before synchronization.
     *  The first communicant gets index 0.
     */

    /**
     * Registers another communicant for possible synchronization.
     * @param newComm   The communicant that could synchronize its data with this object.
     * @param index     The index in the communicant list at which to insert this communicant.
     *                  The order of communicants dictates the order in which they are
     *                  synchronized upon a synchronization call.
     *                  By default, new communicants are added to the head of the list
     */
    void addComm(Communicant* newComm, int index = 0);

    /**
     * Delete all communicants oldComm (i.e. stored at the same memory address) from the communicant list.
     * @param oldComm  A pointer to the desired communicant.
     */
    void delComm(Communicant* oldComm);

    /**
     * Delete the communicant at the given index in the communicant list.
     * @param index  Index of the communicant to delete.
     */
    void delComm(int index);

    /**
     * @return The number of communicants currently registered. 
     */
    int numComm();


    // SYNCHRONIZATION AGENT MANIPULATION
    /* Synchronization agents represent the type of synchronization that could be
     * used to synchronize the data on communicants.  For example, the CPISync
     * synchronization agent uses the CPISync algorithm to synchronize communicant data.
     *  The first agent gets index 0.
     */

    /**
     * Registers another synchronization agent.
     * @param newAgent  The synchronization agent to add
     * @param index     The index in the agent list at which to insert this communicant.
     *                  The order of agents is not significant.
     *                  By default, new agents are added to the head of the list
     */
    void addSyncAgt(SyncMethod* newAgt, int index = 0);

    /**
     * Delete all agents oldAgent (i.e. stored at the same memory address) from the agent list.
     * @param oldAgt  A pointer to the desired synchronization agent.
     */
    void delSyncAgt(SyncMethod* oldAgt);

    /**
     * Delete the agent at the given index in the agent list.
     * @param index  Index of the agent to delete.
     */
    void delSyncAgt(int index);



    // SYNCHRONIZATION METHODS
    /**
     * Sequentially listens for a synchronization request from each communicant.
     * In other words, listens and waits for a synchronization request from the first
     * communicant.  When that request comes through and synchronization is complete,
     * listens for a request from the second communicant ...
     * @param sync_num  The method of synchronization request to listen for.  This is
     *          an index into the list of synchronization methods supplied upon construction.
     *          Thus, if the first synchronization method supplied in the Constructor is
     *          a CPISync method, then sync_num=0 (the default value) will listen for a CPISync sync request.
     * @return true iff all synchronizations were completed successfully
     */
    bool listenSync(int sync_num = 0);

    /**
     * Sequentially sends a specific synchronization request to each communicant.
     * @param sync_num  This is an index into the list of synchronization methods supplied upon construction.
     *          Thus, if the first synchronization method supplied in the constructor is
     *          a CPISync method, then sync_num=0 (the default value) will listen for a CPISync sync request.
     * @return           true iff all synchronizations were completed successfully
     */
    bool startSync(int sync_num);



    // INFORMATIONAL
    /**
     * @return the port on which the server is listening for communicant comm_index.
     * If no server is listening for this communicant, the port returned is -1
     * @param comm_index       The index of the communicant that interests us.
     * @note:  running time is Theta(comm_index) since the underlying data structure is a linked list
     * */
    int getPort(int comm_index);

    /**
     * Displays some internal information about this method
     */
    string info() {
        return "I am a GenSync object";
    }

    /**
     * Destructor
     */
    ~GenSync();

private:
    // METHODS
    /**
     * No argument constructor ... should not be used
     */
    GenSync();


    // FIELDS
    /** A container for the data stored by this GenSync object. */
    list<DataObject*> myData;

    /** A list of communicants registered to be able to sync with this GenSync object. */
    list<Communicant*> myCommList;

    /** A list of synchronization methods that can be used to sync with this GenSync object. */
    list<SyncMethod*> mySyncList;
    
    /** The file to which to output any additions to the data structure. */
    ofstream *outFile;
};

#endif
